# vim: set fileencoding=utf-8 :

# connord - connect to nordvpn servers
# Copyright (C) 2019  Mael Stor <maelstor@posteo.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Manages server categories"""

from connord import ConnordError
from connord.formatter import Formatter

CATEGORIES = {
    "double": "Double VPN",
    "dedicated": "Dedicated IP",
    "standard": "Standard VPN servers",
    "p2p": "P2P",
    "obfuscated": "Obfuscated Servers",
    "onion": "Onion Over VPN",
}


class CategoriesError(ConnordError):
    """Throw within this module"""


def verify_categories(categories):
    """Verify if categories is are valid

    :categories: a list of categories from the command-line
    :raises: CategoriesError if there are invalid categories in categories
    :returns: True if all categories are valid
    """

    if not isinstance(categories, list):
        raise CategoriesError("Wrong server categories: {!s}".format(categories))

    wrong_categories = []
    for server_type in categories:
        if server_type not in CATEGORIES.keys():
            wrong_categories.append(server_type)

    if wrong_categories:
        raise CategoriesError("Wrong server categories: {!s}".format(wrong_categories))

    return True


def verify_categories_description(descriptions):
    """Verify if categories descriptions are valid

    :description: a list of descriptions
    :raises: CategoriesError if there are invalid categories in categories
    :returns: True if all type descriptions are valid
    """

    if not isinstance(descriptions, list):
        raise CategoriesError("Wrong type: {!s}".format(type(descriptions)))

    wrong_categories = [desc for desc in descriptions if desc not in CATEGORIES.values()]

    if wrong_categories:
        raise CategoriesError("Wrong category descriptions: {!s}".format(wrong_categories))

    return True


def map_categories(categories):
    """Map categories from command-line to strings used by nordvpn api.

    :categories: a list of categories from the command-line.
    :returns: a list of mapped categories.
    """

    mapped_categories = [CATEGORIES[category] for category in categories]
    return mapped_categories


def map_categories_reverse(categories):
    """Map categories from descriptions to the internal used type codes."""

    verify_categories_description(categories)
    mapped_categories = [
        key for category in categories for key, value in CATEGORIES.items() if category == value
    ]

    return mapped_categories


def has_category(server, category):
    """Return true if a server has category in categories."""
    for category_ in server["categories"]:
        if category_["name"] == category:
            return True

    return False


def filter_servers(servers, categories=None):
    """Filter a list of servers by type (category).

    :servers: List of servers (parsed from nordvpn api to json).
    :categories: List of categories (categories). If None or empty the default
            'standard' is applied.
    :returns: The filtered list.
    """

    if categories is None or not categories:
        categories = ["standard"]

    mapped_categories = map_categories(categories)

    filtered_servers = []
    servers = servers.copy()
    for server in servers:
        append = True
        for mapped_category in mapped_categories:
            if not has_category(server, mapped_category):
                append = False
                break

        if append:
            filtered_servers.append(server)

    return filtered_servers


class CategoriesPrettyFormatter(Formatter):
    """Format type in pretty format"""

    def format_headline(self, sep="="):
        """Format headline

        :param sep: the fill character
        :returns: centered string
        """

        categories_header = "Categories"
        return self.center_string(categories_header, sep)

    def format_category(self, category, description):
        """Format a category

        :param category: the category
        :param description: the description
        :returns: the formatted category as string
        """

        return "  {:26}{}".format(category, description)


def to_string(stream=False):
    """Gather all categories in a printable string

    :param stream: If True print to stdout else print to formatter.output variable
    :returns: Formatted string if stream is False else an empty string
    """
    formatter = CategoriesPrettyFormatter()
    file_ = formatter.get_stream_file(stream)

    headline = formatter.format_headline()
    print(headline, file=file_)

    for category, description in CATEGORIES.items():
        formatted_category = formatter.format_category(category, description)
        print(formatted_category, file=file_)

    print(formatter.format_ruler(sep="-"), file=file_)
    return formatter.get_output()


def print_categories():
    """Prints all possible categories"""
    to_string(stream=True)



# vim: set fileencoding=utf-8 :

# connord - connect to nordvpn servers
# Copyright (C) 2019  Mael Stor <maelstor@posteo.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""Manage all resources of connord in a centralized module"""

import os
import getpass
import tempfile
from shutil import rmtree
import yaml
from pkg_resources import resource_filename
from connord import ConnordError
from connord import user

__NORDVPN_DIR = "/etc/openvpn/client/nordvpn"
__SCRIPTS_DIR = "/etc/openvpn/client/scripts"
__CONFIG_DIR = "/etc/connord"
__CONFIG_FILE = __CONFIG_DIR + "/config.yml"
__RUN_DIR = "/var/run/connord"

__DATABASE_FILE = resource_filename(__name__, "db/connord.sqlite3")


class ResourceNotFoundError(ConnordError):
    """Raised when a resource is requested but doesn't exist"""

    def __init__(self, resource_file, message=None):

        # if a message is given use it else use the one defined here
        if message:
            super().__init__(message)
        else:
            super().__init__("Resource does not exist: {!r}".format(resource_file))

        #: A string representing a file path
        self.resource_file = resource_file


class MalformedResourceError(ConnordError):
    """Raise when a resource (like yaml) could not be parsed due to parsing errors."""

    def __init__(self, resource_file, problem, problem_mark):
        """Init

        :param resource_file: the file in question
        :param problem: description of the problem
        :param problem_mark: hint where the problem arose
        """
        super().__init__(
            "Malformed resource: {!r}\n{!s}\n{!s}".format(
                resource_file, problem, problem_mark
            )
        )

        self.resource_file = resource_file
        self.problem = problem
        self.problem_mark = problem_mark


def list_dir(directory, filetype=None):
    """Return a file list of a directory with absolute paths. May be filtered by a
    specific filetype (extension).
    """
    files = os.listdir(directory)

    if filetype:
        files = [_file for _file in files if _file.endswith("." + filetype)]

    full_path_files = [directory + "/" + _file for _file in files]

    return full_path_files


def get_zip_dir(create=True):
    """Returns the directory where nordvpn related zip files are stored

    :param create: if True create the directory
    :returns: the zip directory
    :rtype: string
    :raises: ``ResourceNotFoundError`` when `zip_dir` does not exist
    """

    zip_dir = __NORDVPN_DIR
    if not os.path.exists(zip_dir):
        if create:
            os.makedirs(zip_dir)
        else:
            raise ResourceNotFoundError(zip_dir)

    return zip_dir


def get_zip_path(zip_name=None):
    """Return the absolute path to `zip_name` with the output of the
    ``get_zip_dir()`` as base directory. Does not check if the path
    exists.

    :param zip_name: name of the zip file
    :ptype: string
    :default: None
    :returns: the path to the zip file
    :rtype: string
    """

    if not zip_name:
        zip_name = "ovpn.zip"

    zip_dir = get_zip_dir(create=False)
    return zip_dir + "/" + zip_name


def get_zip_file(zip_name=None, create_dirs=True):
    """Return the absolute path to `zip_name` with the output of the
    ``get_zip_dir()`` as base directory. Verifies if the file exists.

    :param zip_name: name of the zip file
    :ptype: string
    :default: None
    :returns: the path to the zip file
    :rtype: string
    :raises: ResourceNotFoundError if the `zip_path` does not exist
    """

    if not zip_name:
        zip_name = "ovpn.zip"

    zip_dir = get_zip_dir(create=create_dirs)
    zip_path = zip_dir + "/" + zip_name
    if os.path.exists(zip_path):
        return zip_path

    raise ResourceNotFoundError(zip_path)


def get_database_file(create=True):
    """Return the database file path

    :param create: if True create the database file (not the directory)
    :returns: the file path as string
    """

    database_file = __DATABASE_FILE
    if not os.path.exists(database_file):
        if create:
            with open(database_file, "w"):
                pass
        else:
            raise ResourceNotFoundError(database_file)

    return database_file


def file_has_permissions(path, permissions=0o600):
    """Check file permissions"""

    stats = os.stat(path)
    return stats.st_mode & 0o777 == permissions


def verify_file_permissions(path, permissions=0o600):
    """Verify file permissions"""

    if not file_has_permissions(path, permissions):
        raise PermissionError(
            "Unsafe file permissions: {!r} should have mode: {!r}.".format(
                path, oct(permissions)
            )
        )
    return True


def get_credentials_dir(create=True):
    """Return the directory where the default credentials file can be found"""
    creds_dir = __NORDVPN_DIR
    if not os.path.exists(__NORDVPN_DIR):
        if create:
            os.makedirs(__NORDVPN_DIR, mode=0o755)
        else:
            raise ResourceNotFoundError(__NORDVPN_DIR)

    return creds_dir


@user.needs_root
def get_credentials_file(file_name="credentials", create=True):
    """Return the credentials file path"""
    creds_dir = get_credentials_dir()
    creds_file = "{}/{}".format(creds_dir, file_name)
    if not os.path.exists(creds_file):
        if create:
            create_credentials_file(creds_file)
        else:
            raise ResourceNotFoundError(creds_file)

    verify_file_permissions(creds_file)
    return creds_file


def _get_username():
    """Ask for the username in the terminal and return the value"""
    return input("Enter username: ")


def create_credentials_file(credentials_file):
    """Create the credentials file from user input. Returns the resulting path"""
    username = _get_username()
    password = getpass.getpass("Enter password: ")

    with open(credentials_file, "w") as creds_fd:
        creds_fd.write(username + "\n")
        creds_fd.write(password + "\n")

    password = None
    os.chmod(credentials_file, mode=0o600)
    return credentials_file


@user.needs_root
def get_ovpn_dir(create=True):
    """Return the directory path where the ovpn_udp and ovpn_tcp directories can be
    found. Raises a ResourceNotFoundError if the directory does not exist.
    """
    if os.path.exists(__NORDVPN_DIR):
        return __NORDVPN_DIR

    raise ResourceNotFoundError(__NORDVPN_DIR)


@user.needs_root
def get_ovpn_protocol_dir(protocol="udp"):
    """Return either the ovpn_udp or ovpn_tcp directory path.

    :raises: ResourceNotFoundError if the directory does not exist.
    """
    base_dir = get_ovpn_dir()
    config_dir = "{}/ovpn_{}".format(base_dir, protocol)

    if os.path.exists(config_dir):
        return config_dir

    raise ResourceNotFoundError(config_dir)


@user.needs_root
def get_ovpn_config(domain, protocol="udp"):
    """Returns the configuration file of the corresponding domain with protocol either
    'udp' or 'tcp'

    :raises: ResourceNotFoundError if the file does not exist
    """
    config_dir = get_ovpn_protocol_dir(protocol)
    if ".nordvpn.com" not in domain:
        domain = "{}.nordvpn.com".format(domain)

    config_file = "{}/{}.{}.ovpn".format(config_dir, domain, protocol)
    if os.path.exists(config_file):
        return config_file

    raise ResourceNotFoundError(config_file)


def get_scripts_dir():
    """Return the directory where the scripts are stored."""

    if os.path.exists(__SCRIPTS_DIR):
        scripts_dir = __SCRIPTS_DIR
    else:
        scripts_dir = resource_filename(__name__, "scripts")

    return scripts_dir


def get_scripts_file(script_name="openvpn_up_down.bash"):
    """Return an absolute file path to script_name.

    :raises: ResourceNotFoundError if the file not exists.
    """

    scripts_dir = get_scripts_dir()
    scripts_file = "{}/{}".format(scripts_dir, script_name)
    if os.path.exists(scripts_file):
        return scripts_file

    raise ResourceNotFoundError(scripts_file)


def get_config_dir():
    """Return the directory path to configuration files"""
    if os.path.exists(__CONFIG_DIR):
        config_dir = __CONFIG_DIR
    else:
        config_dir = resource_filename(__name__, "config")

    return config_dir


# TODO: has_config(filetype=None)


def list_config_dir(filetype=None):
    """Returns a file list of the configuration directory. May be filtered by
    filetype."""
    config_dir = get_config_dir()
    return list_dir(config_dir, filetype)


def get_config_file(config_name="config.yml"):
    """Return a configuration file from the configuration directory

    :raises: ResourceNotFoundError if the file does not exist.
    """

    config_dir = get_config_dir()
    config_file = "{}/{}".format(config_dir, config_name)
    if os.path.exists(config_file):
        return config_file

    raise ResourceNotFoundError(config_file)


# TODO: rename to read_config
def get_config():
    """Returns a dictionary parsed from a yaml configuration file

    :raises: MalformedResourceError in case of parsing errors
    """
    config_file = get_config_file()
    try:
        with open(config_file, "r") as config_fd:
            return yaml.safe_load(config_fd)
    except yaml.parser.ParserError as error:
        raise MalformedResourceError(
            config_file, error.problem, str(error.problem_mark)
        )


def write_config(config_dict):
    """Write to the yaml configuration file (config.yml) parsed from config_dict.

    :raises: TypeError if config_dict is not a dictionary
    """
    config_file = get_config_file()
    if not isinstance(config_dict, dict):
        raise TypeError(
            # pylint: disable=line-too-long
            "Could not write to {!r}: Invalid type: Found {!s} but expected {!s}.".format(
                config_file, type(config_dict), dict
            )
        )

    with open(config_file, "w") as config_fd:
        yaml.dump(config_dict, config_fd, default_flow_style=False)


@user.needs_root
def get_stats_dir(create=True):
    """Return path to the stats directory.

    :raises: ResourceNotFoundError if path doesn't exist and create is false.
    """
    stats_dir = __RUN_DIR
    if not os.path.exists(stats_dir):
        if create:
            os.makedirs(stats_dir, mode=0o750)
        else:
            raise ResourceNotFoundError(stats_dir)

    return stats_dir


@user.needs_root
def remove_stats_dir():
    """Delete the content of the stats directory. Does nothing if it doesn't exists."""
    try:
        stats_dir = get_stats_dir(create=False)
        rmtree(stats_dir)
    except ResourceNotFoundError:
        pass

    return True


def list_stats_dir(filetype=None):
    """Return the content of the stats directory. Files can be filtered by filetype.
    """
    stats_dir = get_stats_dir()
    return list_dir(stats_dir, filetype)


@user.needs_root
def get_stats_file(stats_name=None, create=True):
    """Return the path to the stats_name file.

    :raises: ResourceNotFoundError if the file doesn't exist and create is false.
    """
    if not stats_name:
        stats_name = "stats"

    stats_dir = get_stats_dir(create)
    stats_file = "{}/{}".format(stats_dir, stats_name)
    if not os.path.exists(stats_file):
        if create:
            with open(stats_file, "w"):
                pass
        else:
            raise ResourceNotFoundError(stats_file)

    return stats_file


# TODO: Rename to read_stats
@user.needs_root
def get_stats(stats_name="stats"):
    """Return a dictionary parsed from a yaml stats file.

    :raises: MalformedResourceError if errors occurred during parsing.
    """
    stats_file = get_stats_file(stats_name=stats_name)
    try:
        with open(stats_file, "r") as stats_fd:
            stats_dict = yaml.safe_load(stats_fd)
            if not stats_dict:
                stats_dict = {}

            return stats_dict
    except yaml.parser.ParserError as error:
        raise MalformedResourceError(stats_file, error.problem, str(error.problem_mark))


@user.needs_root
def write_stats(stats_dict, stats_name="stats"):
    """Write a dictionary to a stats_file name stats_name.

    :raises: TypeError if stats_dict is not of type dict
    """
    stats_file = get_stats_file(stats_name=stats_name)
    if not isinstance(stats_dict, dict):
        raise TypeError(
            # pylint: disable=line-too-long
            "Could not write to {!r}: Invalid type: Found {!s} but expected {!s}.".format(
                stats_file, type(stats_dict), dict
            )
        )

    with open(stats_file, "w") as stats_fd:
        yaml.dump(stats_dict, stats_fd, default_flow_style=False)


def read_pid(pid_name="openvpn.pid"):
    """Return the content of a pid file as integer. Pid files reside in stats_dir.
    """
    pid_file = get_stats_file(stats_name=pid_name)
    with open(pid_file, "r") as pid_fd:
        return int(pid_fd.readline())


def get_ovpn_tmp_path():
    """Return the path to the temporary file to be used for temporarily
    created ovpn files.
    """
    temp_dir = tempfile.gettempdir()
    return "{}/{}".format(temp_dir, "ovpn.conf")


def remove_ovpn_tmp_file():
    """Delete the temporary file used for ovpn."""
    temp_dir = tempfile.gettempdir()
    os.remove("{}/{}".format(temp_dir, "ovpn.conf"))



# vim: set fileencoding=utf-8 :

# connord - connect to nordvpn servers
# Copyright (C) 2019  Mael Stor <maelstor@posteo.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""Manage location database and formatting of areas"""

import time
import requests
from cachetools import cached, LRUCache
import cachetools.func
from connord import ConnordError, Printer
from connord import servers
from connord import sqlite
from connord.formatter import Formatter


class AreaError(ConnordError):
    """Thrown within this module"""


API_URL = "https://nominatim.openstreetmap.org"


@cached(cache=LRUCache(maxsize=50))
def query_location(latitude, longitude):
    """Query location given with latitude and longitude coordinates
    from remote nominatim api. The values are cached to reduce queries.
    The nominatim api restricts queries to 1/sec.


    :param latitude: string with the latitude in float notation
    :param longitude: string with the longitude in float notation
    :returns: dictionary of the response in json
    """
    header = {
        "User-Agent": "Mozilla/5.0 (Windows NT 6.1; Win64; x64; rv:60.0) "
        "Gecko/20100101 Firefox/60.0"
    }

    endpoint = "reverse"
    flags = {
        "lat": latitude,
        "lon": longitude,
        "format": "jsonv2",
        "addressdetails": "1",
        "accept-language": "en",
        "zoom": "18",
    }
    url = "{}/{}.php?".format(API_URL, endpoint)
    for k, v in flags.items():
        url += "{}={}&".format(k, v)

    url = url.rstrip("&")
    with requests.get(url, headers=header, timeout=1) as response:
        time.sleep(1)
        return response.json()


def init_database(connection):
    """Initialize location database

    :param connection: a database connection
    """

    with connection:
        sqlite.create_location_table(connection)


def update_database():
    """Updates the location database with least possible online queries."""
    connection = sqlite.create_connection()
    with connection:
        init_database(connection)

    servers_ = servers.get_servers()
    printer = Printer()
    progress_bar = printer.incremental_bar(
        "Updating location database", max=len(servers_)
    )

    for server in servers_:
        longitude = str(server["location"]["long"])
        latitude = str(server["location"]["lat"])

        connection = sqlite.create_connection()
        with connection:
            if sqlite.location_exists(connection, latitude, longitude):
                progress_bar.next()
                continue
            else:
                location_d = query_location(latitude, longitude)

        display_name = location_d["display_name"]
        city_keys = ["city", "town", "village", "residential", "state"]
        for key in city_keys:
            try:
                city = location_d["address"][key]
                break
            except KeyError:
                continue
        else:
            city = "Unknown"

        country = server["country"]
        country_code = server["flag"].lower()
        location = (latitude, longitude, display_name, city, country, country_code)

        connection = sqlite.create_connection()
        with connection:
            sqlite.create_location(connection, location)

        # if not printer.quiet:
        progress_bar.next()

    progress_bar.finish()


def get_server_angulars(server):
    """Return latitude and longitude from server

    :param server: a dictionary describing a server
    :returns: tuple with latitude and longitude
    """
    latitude = str(server["location"]["lat"])
    longitude = str(server["location"]["long"])
    return (latitude, longitude)


def verify_areas(areas_):
    """Verify a list of areas if they can be resolved to valid areas saved in
    the location database. Ambiguous strings resolve to more than one location.

    :param areas_: list of areas
    :returns: list of found areas
    :raises: ValueError if an area could not be found
             AreaError if the area string is ambiguous
    """
    if not isinstance(areas_, list):
        raise TypeError(
            "Expected areas to be <class 'list'>: But found {!s}".format(type(areas_))
        )

    locations = get_locations()
    translation_table = get_translation_table()

    areas_not_found = []
    # side effect: get rid of double entries in areas_ from command-line
    areas_found = {area: list() for area in areas_}
    for area in areas_found.keys():
        for location in locations:
            city = location["city"]
            city_trans = city.translate(translation_table)
            city_lower = city_trans.lower()
            if city_lower.startswith(area):
                areas_found[area].append(city)

        if not areas_found[area]:
            areas_not_found.append(area)

    if areas_not_found:
        raise ValueError("Areas not found: {!s}".format(areas_not_found))

    ambiguous_areas = {}
    for area, cities in areas_found.items():
        if len(cities) > 1:
            ambiguous_areas[area] = cities

    if ambiguous_areas:
        error_string = ""
        for area, cities in ambiguous_areas.items():
            error_string += " {}: {},".format(area, cities)

        error_string = error_string.rstrip(",")
        raise AreaError("Ambiguous Areas:{}".format(error_string))

    return [area for area in areas_found.keys()]


def get_translation_table():
    """Translate special characters to the english equivalent

    :returns: a translation table
    """
    return str.maketrans("áãčëéşșť", "aaceesst")


def filter_servers(servers_, areas_):
    """Filter servers by areas

    :param servers_: list of servers as dictionary
    :param areas_: list of areas as string
    :returns: servers which match the list of areas
    :raises: TypeError when servers_ is None
    """

    if servers_ is None:
        raise TypeError("Servers may not be None")

    if areas_ is None or not areas_ or not servers_:
        return servers_

    translation_table = get_translation_table()
    areas_lower = [str.lower(area) for area in areas_]
    areas_trans = [area.translate(translation_table) for area in areas_lower]

    filtered_servers = []
    connection = sqlite.create_connection()
    with connection:
        for server in servers_:
            lat, lon = get_server_angulars(server)
            if not sqlite.location_exists(connection, lat, lon):
                update_database()

            city = sqlite.get_city(connection, lat, lon)[0]
            city = city.translate(translation_table)
            city = city.lower()
            for area in areas_trans:
                if city.startswith(area):
                    filtered_servers.append(server)
                    break

    return filtered_servers


def get_min_id(city):
    """Calculate the minimum string which must be given to identify an area
    unambiguously

    :param city: the area/city as string
    :returns: the minimum string
    """

    min_id = ""
    translation_table = get_translation_table()
    for char in city:
        char = char.translate(translation_table)
        min_id += char.lower()
        try:
            verify_areas([min_id])
            break
        except AreaError:
            continue

    return min_id


@cachetools.func.ttl_cache(ttl=60, maxsize=1)
def get_locations():
    """Return all locations found in the database. If the database does not exist
    update the database

    :returns: a list of all locations
    """
    connection = sqlite.create_connection()
    with connection:
        locations = sqlite.get_locations(connection)
        if not locations:
            update_database()
            return sqlite.get_locations(connection)

        return locations


class AreasPrettyFormatter(Formatter):
    """Format areas in pretty format"""

    def format_headline(self, sep="="):
        """Format the headline

        :param sep: filling character and separator
        :returns: the headline
        """
        headline = self.format_ruler(sep) + "\n"
        headline += "{:8}: {:^15} {:^15}  {:4}\n".format(
            "Mini ID", "Latitude", "Longitude", "City"
        )
        headline += "{}\n".format("Address")
        headline += self.format_ruler(sep)
        return headline

    def format_area(self, location):
        """Format the area

        :param location: tuple of strings (latitude, longitude)
        :returns: the area as string
        """
        lat = float(location["latitude"])
        lon = float(location["longitude"])
        display_name = location["display_name"]
        city = location["city"]
        min_id = get_min_id(city)

        string = "{!r:8}: {: 14.9f}° {: 14.9f}°  {:40}\n".format(min_id, lat, lon, city)
        string += "{}\n".format(display_name)
        string += self.format_ruler(sep="-")
        return string


def to_string(stream=False):
    """High-level command to format areas and returns the resulting string if not
    streaming directly to screen.

    :param stream: True if the output shall be streamed to stdout
    :returns: When streaming an empty string or else the result of the formatter
    """
    formatter = AreasPrettyFormatter()
    file_ = formatter.get_stream_file(stream)

    locations = get_locations()

    headline = formatter.format_headline()
    print(headline, file=file_)

    for location in locations:
        area = formatter.format_area(location)
        print(area, file=file_)

    return formatter.get_output()


def print_areas():
    """High-level command to print areas to stdout"""
    to_string(stream=True)



# vim: set fileencoding=utf-8 :

# connord - connect to nordvpn servers
# Copyright (C) 2019  Mael Stor <maelstor@posteo.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""An openvpn wrapper to connect to NordVPN servers"""


import subprocess
from multiprocessing.pool import ThreadPool
from multiprocessing import cpu_count
from math import inf
import time
import os
import re
import signal
from connord import ConnordError, Printer
from connord import iptables
from connord import servers
from connord import load
from connord import countries
from connord import areas
from connord import categories
from connord import features
from connord import resources
from connord import update


class ConnectError(ConnordError):
    """Thrown within this module"""


def ping(server):
    """
    Ping a server
    :param dict server: A server as dictionary
    :returns: copy of server with additional key 'ping'
    """

    server_copy = server.copy()
    pattern = re.compile(r"rtt .* = ([\d\.]+)/([\d\.]+)/([\d\.]+)/.* ms")
    ip_address = server["ip_address"]
    with subprocess.Popen(
        ["ping", "-q", "-n", "-c", "1", "-l", "1", "-W", "1", ip_address],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    ) as ping_:

        out, _ = ping_.communicate()
        mat = pattern.search(out.decode())
        if mat:
            server_copy["ping"] = float(mat.group(2))
        else:
            server_copy["ping"] = float("inf")

        printer = Printer()
        printer.info(
            "{:6}: ping: {}".format(server_copy["domain"], server_copy["ping"])
        )
        return server_copy


def ping_servers_parallel(servers_):
    """
    Ping a list of servers
    :param list servers_: List of servers
    :returns: List of servers with ping time
    """
    worker_count = cpu_count() + 1
    with ThreadPool(processes=worker_count) as pool:
        results = []
        for server in servers_:
            results.append(pool.apply_async(ping, (server,)))

        pinged_servers = []
        for result in results:
            pinged_servers.append(result.get())

        return pinged_servers


# pylint: disable=too-many-arguments
def filter_servers(
    servers_, netflix, countries_, areas_, features_, categories_, load_, match
):
    """Filter servers by keys from command-line options
    :param servers_: a list of servers given as dictionary
    :param netflix: Boolean if servers should be filtered to netflix optimized servers
    :param countries_: a list of countries each given as string
    :param areas_: a list of areas each given as string
    :param features_: a list of features each given as string
    :param categories_: a list of categories each given as string
    :param load_: depending on match filter max, min or equal load servers
    :param match: may be 'max', 'min' or 'equal'
    :returns: the filtered list of servers
    """
    servers_ = servers_.copy()
    if load_:
        servers_ = load.filter_servers(servers_, load_, match)
    if netflix:
        servers_ = servers.filter_netflix_servers(servers_, countries_)
    if countries_:
        servers_ = countries.filter_servers(servers_, countries_)
    if areas_:
        servers_ = areas.filter_servers(servers_, areas_)
    if categories_:
        servers_ = categories.filter_servers(servers_, categories_)
    if features_:
        servers_ = features.filter_servers(servers_, features_)

    return servers_


def filter_best_servers(servers_):
    """Filter servers by lowest load and ping

    :param servers_: list of servers each given as dictionary
    :returns: the filtered list of servers
    """
    servers_ = servers_.copy()
    servers_ = sorted(servers_, key=lambda k: k["load"])
    if len(servers_) > 10:
        servers_ = servers_[:10]
    servers_ = ping_servers_parallel(servers_)
    servers_ = sorted(servers_, key=lambda k: k["ping"])
    return servers_


def connect_to_specific_server(domain, openvpn, daemon, protocol):
    """Connect to a specific server

    :param domain: list of domains which holds one value like gb111
    :param openvpn: string of options to pass-through to openvpn
    :param daemon: True if the openvpn shall be run in daemon mode
    :param protocol: may be one of 'udp' or 'tcp'
    :returns: True if openvpn was run successfully
    """
    server = servers.get_server_by_domain(domain[0])
    printer = Printer()
    if categories.has_category(server, "Obfuscated Servers"):
        if not printer.yes_no(
            (
                "WARNING: {} is an obfuscated server.\n"
                "This may fail if not configured properly.\n"
                "Are you sure you want to continue?"
            ).format(server["domain"])
        ):
            return False
    return run_openvpn(server, openvpn, daemon, protocol)


# pylint: disable=too-many-locals
def connect(
    domain,
    countries_,
    areas_,
    features_,
    categories_,
    netflix,
    load_,
    match,
    daemon,
    openvpn,
    protocol,
):
    """High-level function to connect to a openvpn server. Filters servers and
    tries to connect to an openvpn server 'max_retries' times or else raise an
    exception.

    :param domain: list of domains. defaults to one value 'best'.
    :param countries_: list of countries
    :param areas_: list of areas
    :param features_: list of features
    :param categories_: list of categories
    :param netflix: True to filter netflix optimized servers
    :param load_: depending on match, filter by max, min or equal load servers
    :param match: may be 'max', 'min' or 'equal'
    :param daemon: True if openvpn shall run in daemon mode
    :param config_: the path to the openvpn configuration file
    :param openvpn: options to pass-through to openvpn as string
    :param protocol: may be 'udp' or 'tcp'
    :returns: True if running openvpn was successful
    :raises: ConnectError
    """

    if "best" not in domain:
        return connect_to_specific_server(domain, openvpn, daemon, protocol)

    servers_ = servers.get_servers()
    servers_ = filter_servers(
        servers_, netflix, countries_, areas_, features_, categories_, load_, match
    )

    best_servers = filter_best_servers(servers_)
    max_retries = 3
    printer = Printer()
    for i, server in enumerate(best_servers):
        if i == max_retries:
            raise ConnectError("Maximum retries reached.")

        if categories.has_category(server, "Obfuscated Servers"):
            if not printer.yes_no(
                (
                    "WARNING: {} is an obfuscated server.\n"
                    "This may fail if not configured properly.\n"
                    "Are you sure you want to continue?"
                ).format(server["domain"])
            ):
                continue

        if server["ping"] != inf:
            printer.info("Trying to connect to {}".format(server["domain"]))
            if run_openvpn(server, openvpn, daemon, protocol):
                return True

            # else give the next server a try
        else:
            raise ConnectError("No server left with a valid ping.")

    raise ConnectError("No server found to establish a connection.")


# TODO: Rename to OpenvpnCommandPanicError
class OpenvpnCommandPanic(ConnectError):
    """Raised when something went wrong running openvpn"""

    def __init__(self, problem, message=None):
        """Initialize

        :param problem: string to describe what went wrong
        :param message: If not None apply default message.
        """
        if not message:
            message = "Running openvpn failed: {}".format(problem)

        super().__init__(message)
        self.problem = problem


class OpenvpnCommand:
    """Wraps openvpn"""

    def __init__(self, server, openvpn, daemon, protocol):
        """ Init

        :param server: a server dictionary
        :param openvpn: openvpn options as one string
        :param daemon: True if openvpn shall be run in daemon mode
        :param protocol: 'udp' or 'tcp'
        """
        self.server = server
        self.domain = server["domain"]
        self.openvpn_options = openvpn
        self.daemon = daemon
        self.protocol = protocol
        self.cmd = ["openvpn"]

    def _add_openvpn_cmd_option(self, flag, *args):
        """Convenience function to add an option to the command list

        :param flag: the --flag
        :param args: possible arguments for flag
        """
        if flag.startswith("-"):
            if flag not in self.cmd:
                self.cmd.append(flag)
                for arg in args:
                    self.cmd.append(arg)
        else:  # flag seems to be an argument and can be added without checks
            self.cmd.append(flag)

    @staticmethod
    def _format_flag(flag):
        """Return a formatted flag

        :param flag: an option without the necessary leading '--'
        """
        return "--{}".format(flag)

    def _forge_number(self, key, value):
        """Converts a number value to string and adds the option to self.cmd

        :param key: the flag
        :param value: the number value
        """
        flag = self._format_flag(key)
        self._add_openvpn_cmd_option(flag, str(value))

    def _forge_string(self, key, value):
        """Handles string values and adds the option to self.cmd. The key
        auth-user-pass needs special treatment since the value can be 'built-in'
        to use the predefined path to the credentials file.

        :param key: the flag
        :param value: the value related to the flag
        """
        flag = self._format_flag(key)
        if key == "auth-user-pass":
            if value == "built-in":
                creds_file = resources.get_credentials_file()
            else:
                creds_file = value

            self._add_openvpn_cmd_option(flag, creds_file)
        else:
            self._add_openvpn_cmd_option(flag, value)

    def _forge_bool(self, key, value):
        """Handles boolean values to be added to the final command.

        :param key: the flag
        :param value: the value related to the flag
        """
        if value in (True, "True", "true"):
            flag = self._format_flag(key)
            self._add_openvpn_cmd_option(flag)

    def _forge_list(self, key, list_):
        """Handles list values to be added to the final command. The 'scripts' key
        needs special treatment for the 'built-in' value.

        :param key: the flag
        :param list_: the list of strings
        """
        if key == "scripts":
            self._forge_scripts(list_)
        else:
            flag = self._format_flag(key)
            if list_:
                self._add_openvpn_cmd_option(flag)
                for value in list_:
                    self._add_openvpn_cmd_option(value)
            else:
                self._add_openvpn_cmd_option(flag)

    @staticmethod
    def _format_script_arg(script_name, path, file_):
        """Format the argument for the scripts key to combine the script path and the
        path to the environment file in one string. Handles the special value 'built-in'
        to resolve to the built-in path of the corresponding script.
        :param script_name: name of the script
        :param path: either 'built-in' or a user-defined path
        :param file_: The name for the file created by the script
        :returns: the formatted argument.
        """
        if path == "built-in":
            script_path = resources.get_scripts_file(script_name=script_name)
            env_dir = resources.get_stats_dir()
            env_file = "{}/{}".format(env_dir, file_)
        else:
            script_path = path
            env_file = file_

        return "'{}' {}".format(script_path, env_file)

    def _forge_scripts(self, scripts):
        """Adds the formatted script paths to the resulting command-line

        :param scripts: list of scripts served as dictionaries
        """
        for script in scripts:
            name = script["name"]
            flag = self._format_flag(name)
            path = script["path"]
            file_ = script["creates"]
            if name in ("up", "down"):
                arg = self._format_script_arg("openvpn_up_down.bash", path, file_)
            elif name == "ipchange":
                arg = self._format_script_arg("openvpn_ipchange.bash", path, file_)
            else:
                if path == "built-in":
                    raise resources.ResourceNotFoundError(
                        path, "No built-in found for {!r}.".format(name)
                    )

                arg = self._format_script_arg(name, path, file_)

            self._add_openvpn_cmd_option(flag, arg)

    def _forge_command_line(self):
        """Add the command-line arguments relevant for the openvpn command"""
        if self.openvpn_options:
            for option in self.openvpn_options.split():
                self._add_openvpn_cmd_option(option)

        if self.daemon:
            self._add_openvpn_cmd_option("--daemon")

    def _forge_config(self):
        """Add the openvpn options from 'config.yml' to the openvpn command"""
        openvpn_config = resources.get_config()["openvpn"]
        for k, v in openvpn_config.items():
            if isinstance(v, bool) or v in ("true", "True", "false", "False"):
                self._forge_bool(k, v)
            elif isinstance(v, list):
                self._forge_list(k, v)
            elif isinstance(v, (int, float)):
                self._forge_number(k, v)
            else:
                self._forge_string(k, v)

    def has_flag(self, flag):
        """Return true if the flag is present in the current openvpn command"""
        if flag.startswith("--"):
            return flag in self.cmd

        return "--{}".format(flag) in self.cmd

    def forge_ovpn_config(self, config_file=None):
        """Remove all openvpn command-line arguments from the ovpn config file
        to prevent overriding the command-line. Use final temporary file as
        value for the --config flag

        :param config_file: path to an optional configuration file
        """
        if not config_file:
            config_file = resources.get_ovpn_config(self.domain, self.protocol)

        config_tmp = resources.get_ovpn_tmp_path()

        with open(config_file, "r") as config_fd:
            lines = config_fd.readlines()
        with open(config_tmp, "w") as config_fd:
            for line in lines:
                if line != "\n":
                    # TODO: improve to flag = line.split(None, 1)[0]
                    fake_line = line.rstrip() + " $"
                    flag, _ = fake_line.split(maxsplit=1)
                    if not self.has_flag(flag):
                        config_fd.write(line)
                else:
                    config_fd.write(line)

        os.chmod(config_tmp, mode=0o640)
        self._add_openvpn_cmd_option("--config", config_tmp)

    def forge(self):
        """The high-level command to assemble the openvpn cmd. Adds the --config
        and the --writepid flag flag if not present yet."""
        self._forge_command_line()
        self._forge_config()
        if "--config" not in self.cmd:
            try:
                self.forge_ovpn_config()
            except resources.ResourceNotFoundError:
                update.update(
                    force=True
                )  # give updating a try else let the error happen
                self.forge_ovpn_config()
        else:
            index = self.cmd.index("--config")
            config_file = self.cmd[index + 1]
            self.cmd.remove("--config")
            self.cmd.remove(config_file)
            self.forge_ovpn_config(config_file=config_file)

        if not self.has_flag("--writepid"):
            pid_dir = resources.get_stats_dir(create=True)
            pid_file = pid_dir + "/openvpn.pid"
            self._add_openvpn_cmd_option("--writepid", pid_file)

    def is_daemon(self):
        """Return true if --daemon flag is in the current openvpn command"""
        return "--daemon" in self.cmd

    @staticmethod
    def cleanup():
        """Cleanup temporary stuff and reset iptables."""
        resources.remove_stats_dir()
        iptables.reset(fallback=True)

    @staticmethod
    def is_running(process):
        """Return true if openvpn is running"""
        return not bool(process.poll())

    def panic(self, process, problem):
        """High-level command to shutdown openvpn and cleanup temporary files.

        :raises OpenvpnCommandPanic
        """
        if self.is_running(process):
            process.kill()

        self.cleanup()
        resources.remove_ovpn_tmp_file()
        raise OpenvpnCommandPanic(problem)

    def run(self):
        """High-level command to run openvpn with the assembled command-line.
        Shuts down openvpn after a timeout. Waits this time for openvpn to startup
        and writes the environment files from the scripts. As soon as they are
        present iptables rules are applied. If something goes wrong call the panic
        method

        :returns: True if everything went fine or running in daemon mode.
        """

        config_dict = resources.get_config()["openvpn"]
        self.cleanup()

        printer = Printer()
        printer.info("Running openvpn with '{}'".format(self.cmd))
        with subprocess.Popen(self.cmd) as ovpn:
            # give openvpn a maximum of 60 seconds to startup. A lower value is bad if
            # asked for username/password.
            # pylint: disable=unused-variable
            for i in range(300):
                try:
                    if self.is_running(ovpn):
                        # delay initialization of iptables until resource files are
                        # created. If none are created the delay still applies as normal
                        # timeout
                        time.sleep(0.2)
                        for script in config_dict["scripts"]:
                            stage = script["stage"]
                            if stage in ("up", "always"):
                                resources.get_stats_file(
                                    stats_name=script["creates"], create=False
                                )

                    else:
                        self.panic(ovpn, "Openvpn process stopped unexpected.")

                    if iptables.apply_config_dir(self.server, self.protocol):
                        resources.write_stats(self.server, stats_name="server")

                        stats_dict = resources.get_stats()
                        stats_dict["last_server"] = {}
                        stats_dict["last_server"]["domain"] = self.domain
                        stats_dict["last_server"]["protocol"] = self.protocol
                        resources.write_stats(stats_dict)
                    else:
                        self.panic(ovpn, "Applying iptables failed.")

                    break
                except resources.ResourceNotFoundError:
                    pass
            ### for
            else:
                self.panic(ovpn, "Timeout reached.")

            if self.is_running(ovpn):
                ovpn.wait()

        return True


def run_openvpn(server, openvpn, daemon, protocol):
    """Intermediate function to setup openvpn and wrap the final run."""
    openvpn_cmd = OpenvpnCommand(server, openvpn, daemon, protocol)
    openvpn_cmd.forge()

    retval = False
    try:
        retval = openvpn_cmd.run()
    except KeyboardInterrupt:
        retval = True
        time.sleep(1)

    if not openvpn_cmd.is_daemon():
        resources.remove_ovpn_tmp_file()
        openvpn_cmd.cleanup()

    return retval


def kill_openvpn(pid=None):
    """Kill all openvpn processes currently running or if pid is given this process.
    :param pid: integer with an process id. Can be any pid but intention is to shutdown
    openvpn this way. If None all openvpn process are shutdown targets
    """
    if pid:
        os.kill(pid, signal.SIGTERM)
    else:
        cmd = ["ps"]
        cmd.append("-A")
        with subprocess.Popen(cmd, stdout=subprocess.PIPE) as proc:
            out, _ = proc.communicate()
            for line in out.decode().splitlines():
                if "openvpn" in line:
                    pid = int(line.split(None, 1)[0])
                    os.kill(pid, signal.SIGTERM)



# vim: set fileencoding=utf-8 :

# connord - connect to nordvpn servers
# Copyright (C) 2019  Mael Stor <maelstor@posteo.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
Prints the version along with the copyright.
"""

from connord import __version__, __copyright__, Printer
from connord import resources


def print_version():
    """Prints the version of connord along with the copyright.
    """
    printer = Printer()
    print("connord {}".format(__version__), file=printer)
    print(__copyright__, file=printer)
    printer.info("\nConfiguration directory: '{}'".format(resources.get_config_dir()))



# vim: set fileencoding=utf-8 :

# connord - connect to nordvpn servers
# Copyright (C) 2019  Mael Stor <maelstor@posteo.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import requests
from cachetools import cached, TTLCache
from connord import ConnordError
from connord.categories import map_categories_reverse
from connord import countries
from connord.formatter import Formatter

__API_URL = "https://api.nordvpn.com/server"
NETFLIX = ["us", "ca", "jp", "de", "gb", "fr", "it"]


class DomainNotFoundError(ConnordError):
    """Raised when a domain is requested but doesn't exist."""

    def __init__(self, domain, message=None):
        if not message:
            message = "Domain not found: {!r}".format(domain)

        super().__init__(message)
        self.domain = domain


class MalformedDomainError(ConnordError):
    """Raised when domain is not in the expected format."""

    def __init__(self, domain, problem, message=None):
        if not message:
            message = "Invalid domain: {!r}: {}".format(domain, problem)

        super().__init__(message)
        self.problem = problem


def get_server_by_domain(domain):
    """Return server specified with domain as string. Queries the nordvpn api.

    :raises: ValueError if the domain can't be found
    """
    servers = get_servers()
    if ".nordvpn.com" in domain:
        fqdn = domain
    else:
        fqdn = domain + ".nordvpn.com"

    for server in servers:
        if server["domain"] == fqdn:
            return server

    raise ValueError("Domain not found: {!r}.".format(domain))


def get_servers_by_domains(domains):
    """Abstraction of get_server_by_domain for a list of domains.

    :returns: list of filtered servers
    :raises: DomainNotFoundError if the domain doesn't exist.
    """
    fqdns = []
    for domain in domains:
        if ".nordvpn.com" in domain:
            fqdn = domain
        else:
            fqdn = domain + ".nordvpn.com"

        fqdns.append(fqdn)

    servers = get_servers()
    filtered_servers = []
    for server in servers:
        for fqdn in fqdns:
            if server["domain"] == fqdn:
                filtered_servers.append(server)
                fqdns.remove(fqdn)
                break

    if not fqdns:
        raise DomainNotFoundError(fqdns[0])

    return filtered_servers


@cached(cache=TTLCache(maxsize=1, ttl=60))
def get_servers():
    """Returns the queried servers from nordvpn's api as list of dictionaries."""
    header = {
        "User-Agent": " ".join(
            (
                "Mozilla/5.0 (Windows NT 6.1; Win64; x64; rv:60.0)",
                "Gecko/20100101 Firefox/60.0",
            )
        )
    }

    with requests.get(__API_URL, headers=header, timeout=1) as response:
        return response.json()


def filter_netflix_servers(servers, countries_):
    """Return a list of netflix optimized servers"""
    servers = servers.copy()
    servers = countries.filter_servers(servers, NETFLIX)
    return servers


class ServersPrettyFormatter(Formatter):
    """Class to format servers pretty."""

    def format_headline(self, sep="="):
        """Return formatted headline"""

        headline = self.format_ruler(sep) + "\n"
        headline += "      {:25}  {:6}  {:15}  {:>9}  {}\n".format(
            "Country", "Domain", "IP Address", "Load", "Type"
        )
        headline += "      {}\n".format("Features")
        headline += self.format_ruler(sep)
        return headline

    def format_server(self, server, count, sep="-"):
        """Return pretty formatted server on two lines"""
        ident, _, _ = server["domain"].split(".")
        country = server["country"]
        ip = server["ip_address"]
        load = server["load"]

        categories = [category["name"] for category in server["categories"]]
        categories = ",".join(map_categories_reverse(categories))

        features = [
            feature for feature in server["features"] if server["features"][feature]
        ]
        features = ",".join(features)

        string = "{:4d}: {:25}  {:6}  {:15}  load: {:>3d}  {}\n".format(
            count, country, ident, ip, load, categories
        )
        string += "      {}\n".format(features)
        string += self.format_ruler(sep)

        return string


def to_string(servers, stream=False):
    """If stream is True stream the formatted servers to stdout else to the
    formatter.output variable.

    :returns: the formatted string if stream is False else ''
    """
    if not servers:
        return str()

    formatter = ServersPrettyFormatter()
    file_ = formatter.get_stream_file(stream)

    headline = formatter.format_headline()
    print(headline, file=file_)
    count = 1
    for server in servers:
        server_s = formatter.format_server(server, count)
        print(server_s, file=file_)
        count += 1

    return formatter.get_output()



# vim: set fileencoding=utf-8 :

# connord - connect to nordvpn servers
# Copyright (C) 2019  Mael Stor <maelstor@posteo.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
Init file
"""

import sys
import errno
from progress.bar import IncrementalBar
from progress.spinner import Spinner


class ConnordError(Exception):
    """Main Exception class for connord module"""


# pylint: disable=too-few-public-methods
class Borg:
    """Define a borg class"""

    _shared_state = {}

    def __init__(self):
        self.__dict__ = self._shared_state


class Printer(Borg):
    """Generic printer. Every intentional output must go through this printer
    to recognize quiet and verbose mode set on the command-line or in the
    configuration.
    """

    prefix = "connord: "

    def __init__(self, verbose=False, quiet=False):
        """Initialize the printer. The attributes don't change once they are set.

        :param verbose: if True print info messages
        :param quiet: if True suppress error and info messages
        """

        Borg.__init__(self)
        if "verbose" not in self.__dict__.keys():
            self.verbose = verbose
        if "quiet" not in self.__dict__.keys():
            self.quiet = quiet

    def yes_no(self, question):
        reply = input(question + " (y/N): ").lower().strip()
        if not reply:
            return False
        if reply in ("y", "ye", "yes"):
            return True
        if reply in ("n", "no"):
            return False

        return self.yes_no("Invalid answer. Try 'y' or 'n' or enter for No.")

    def error(self, message):
        """Prints errors if not quiet"""
        if not self.quiet:
            print(self.prefix + message, file=sys.stderr)

    def info(self, message, no_prefix=False, no_newline=False):
        """Prints info messages if verbose"""
        if self.verbose and not self.quiet:
            if no_prefix:
                message_prefixed = message
            else:
                message_prefixed = self.prefix + message
            if no_newline:
                print(message_prefixed, end="")
            else:
                print(message_prefixed)

    @staticmethod
    def write(message):
        """Prints messages independently from verbose and quiet settings
        There's no need to call this function directly. Just pass Printer to the
        file attribute of the __builtin__ print method.

        Example: print('something', file=Printer())
        """
        try:
            print(message, end="")
        except IOError as error:
            if error.errno != errno.EPIPE:
                raise

    @staticmethod
    def format_prefix(prefix):
        if len(prefix) < 40:
            return "{!s:40}".format(prefix)

        return "{!s:40} ".format(prefix)

    class Do:
        def __init__(self, message):
            self.printer = Printer()
            self.message = self.printer.format_prefix(message)

        def __enter__(self):
            self.printer.info(self.message, no_newline=True)

        # pylint: disable=redefined-builtin
        def __exit__(self, type, value, traceback):
            if type:
                self.printer.info("Error", no_prefix=True)
            else:
                self.printer.info("Done", no_prefix=True)

    def incremental_bar(self, message="", **kwargs):
        """Return an Incremental bar if verbose else a NullBar(IncrementalBar)
        which does nothing but can be used like the usual IncrementalBar

        :param message: the prefix message
        :param kwargs: kwargs to pass through to the IncrementalBar
        :returns: IncrementalBar if verbose else NullBar
        """
        if self.verbose:
            return IncrementalBar(self.format_prefix(message), bar_prefix="|", **kwargs)

        return self.NullBar()

    def spinner(self, message="", **kwargs):
        """Return a Spinner"""
        if self.verbose:
            return Spinner(self.format_prefix(message), **kwargs)

        return self.NullSpinner()

    class NullSpinner(Spinner):
        """Fake a Spinner doing nothing"""

        # pylint: disable=super-init-not-called
        def __init__(self, message="", **kwargs):
            pass

        def start(self):
            pass

        def update(self):
            pass

        def show(self, *args, **kwargs):
            pass

        def next(self, n=1):
            pass

        def write(self, s):
            pass

        def finish(self):
            pass

    class NullBar(IncrementalBar):
        """Fake an IncrementalBar doing nothing"""

        # pylint: disable=super-init-not-called
        def __init__(self, message="", **kwargs):
            pass

        def start(self):
            pass

        def update(self):
            pass

        def show(self, *args, **kwargs):
            pass

        def next(self, n=1):
            pass

        def write(self, s):
            pass

        def finish(self):
            pass


__version__ = "0.1.4"
__license__ = "GNU General Public License v3 or later (GPLv3+)"
__copyright__ = """connord  Copyright (C) 2019  Mael Stor <maelstor@posteo.de>
This program comes with ABSOLUTELY NO WARRANTY; This is free software, and you
are welcome to redistribute it under certain conditions; See the LICENSE file
shipped with this software for details."""



# vim: set fileencoding=utf-8 :

# connord - connect to nordvpn servers
# Copyright (C) 2019  Mael Stor <maelstor@posteo.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Wrapper around iptables"""

import os
import re
import netaddr
from iptc import Table, Table6
import iptc
import yaml

from jinja2 import Environment, FileSystemLoader

import netifaces
from connord import ConnordError, Printer
from connord import user
from connord import resources
from connord.formatter import Formatter


class IptablesError(ConnordError):
    """Raise within this module"""


def get_config_path(table, fallback=False, ipv6=False):
    """Return rules path of the table"""
    if fallback:
        if ipv6:
            table_regex = re.compile(r"[0-9]*[-]?{}6.fallback".format(table))
        else:
            table_regex = re.compile(r"[0-9]*[-]?{}.fallback".format(table))

        file_list = resources.list_config_dir(filetype="fallback")
    else:
        if ipv6:
            table_regex = re.compile(r"[0-9]*[-]?{}6.rules".format(table))
        else:
            table_regex = re.compile(r"[0-9]*[-]?{}.rules".format(table))

        file_list = resources.list_config_dir(filetype="rules")

    configs_found = []
    for file_ in file_list:
        result = table_regex.search(file_)
        if result:
            configs_found.append(file_)

    if not configs_found:
        if fallback:
            error_message = (
                "Could not find a fallback configuration for '{}' " "table."
            ).format(table)
        else:
            error_message = "Could not find a configuration for '{}' table.".format(
                table
            )
        raise resources.ResourceNotFoundError(table, error_message)

    elif len(configs_found) > 1:
        raise IptablesError("Ambiguous file paths: {}.".format(configs_found))

    return configs_found[0]


def get_table_name(config_file):
    """Return the table name extracted from the filename"""
    table_regex = re.compile(r"[0-9]*[-]?([a-zA-Z]+[6]?).(rules|fallback)")
    base = os.path.basename(config_file)
    result = table_regex.search(base)
    if result:
        return result.group(1)

    filename = os.path.basename(config_file)
    raise IptablesError(
        "Error: {} is not a valid filename for an iptables rules file.".format(filename)
    )


@user.needs_root
def init_table(table_name):
    """Return a table object with table_name if present in iptc.Table[6].ALL else
    raise a TypeError."""
    if table_name.endswith("6"):
        table_name = table_name[:-1]
        for table in Table6.ALL:
            if table_name == table:
                return Table6(table_name)

        raise TypeError("Error: '{}' is not a valid table.".format(table_name))

    for table in Table.ALL:
        if table_name == table:
            return Table(table)

    raise TypeError("Error: '{}' is not a valid table.".format(table_name))


@user.needs_root
def init_table_from_file_name(config_file):
    """Returns a table object with table name extracted from config_file"""
    table_name = get_table_name(config_file)
    return init_table(table_name)


def is_table_v6(table):
    """Return true if the table is ivp6"""
    return isinstance(table, Table6)


def verify_table(table):
    """Return true if the table is a valid table name

    :param table: table as string
    """

    return table in Table.ALL or table in Table6.ALL


def flush_tables(ipv6=False):
    """Flush all tables and apply the default policy ACCEPT to standard tables"""
    printer = Printer()
    with printer.Do("Flushing all tables: ipv6={!r}".format(ipv6)):
        iptc.easy.flush_all(ipv6=ipv6)

    with printer.Do("Setting ACCEPT policy in all chains"):
        policy = iptc.Policy("ACCEPT")
        for table_s in iptc.easy.get_tables(ipv6):
            for chain_s in iptc.easy.get_chains(table_s, ipv6):
                iptc.easy.set_policy(table_s, chain_s, policy=policy, ipv6=ipv6)


def reset(fallback=True):
    """Reset all tables to fallback if True else just flush them"""
    flush_tables()
    flush_tables(ipv6=True)
    if fallback:
        apply_config_dir(filetype="fallback")


def _apply_config(table, config_d, ipv6):
    for chain_s in config_d:
        if not iptc.easy.has_chain(table, chain_s, ipv6=ipv6):
            iptc.easy.add_chain(table, chain_s, ipv6=ipv6)

        try:
            if config_d[chain_s]["policy"] != "None":
                policy = iptc.Policy(config_d[chain_s]["policy"])
                iptc.easy.set_policy(table, chain_s, policy=policy, ipv6=ipv6)
        except KeyError:
            pass

        try:
            for rule_d in config_d[chain_s]["rules"]:
                if iptc.easy.test_rule(rule_d, ipv6=ipv6):
                    try:
                        iptc.easy.add_rule(table, chain_s, rule_d, ipv6=ipv6)
                    except ValueError:
                        raise IptablesError(
                            "Malformed rule: {}\n  in {}.{}".format(
                                rule_d, table, chain_s
                            )
                        )
                else:
                    raise IptablesError(
                        "Malformed rule: {}\n  in {}.{}".format(rule_d, table, chain_s)
                    )
        except KeyError:
            pass


@user.needs_root
def apply_config(config_file, server=None, protocol=None):
    """Apply a configuration to ip[6]tables (depends on the file name)

    :raises: IptablesError if an invalid rule is present. This leaves the
             table intact with rules applied so far.
    """
    table = init_table_from_file_name(config_file)
    table_s = table.name
    is_ipv6 = is_table_v6(table)

    printer = Printer()
    with printer.Do("Flushing table '{}', ipv6={!r}".format(table_s, is_ipv6)):
        iptc.easy.flush_table(table_s, ipv6=is_ipv6)
        policy = iptc.Policy("ACCEPT")
        for chain_s in iptc.easy.get_chains(table_s, ipv6=is_ipv6):
            iptc.easy.set_policy(table_s, chain_s, policy=policy, ipv6=is_ipv6)

    config_base = os.path.basename(config_file)
    with printer.Do("Applying '{}'".format(config_base)):
        config_d = read_config(config_file, server, protocol)
        _apply_config(table_s, config_d, ipv6=is_ipv6)

    return True


@user.needs_root
def apply_config_dir(server=None, protocol=None, filetype="rules"):
    """High-level command to apply the whole configuration directory with rules or
    fallback files in it.

    :param server: If None this applies 0.0.0.0/0 instead
    :param protocol: If None the default 'udp' is taken
    :param filetype: default is rules but may be fallback too.
    :returns: True on success
    """
    config_files = resources.list_config_dir(filetype=filetype)
    retval = True
    for config_file in config_files:
        if not apply_config(config_file, server, protocol):
            retval = False

    return retval


@user.needs_root
def merge_environment(config_data_dict=None):
    """Merge environment files with the configuration file

    Configuration file variables overwrite variables from environment
    """
    env_dict = dict()

    files = resources.list_stats_dir(filetype="env")
    for file_ in files:
        with open(file_, "r") as file_fd:
            yaml_dict = yaml.safe_load(file_fd)
            env_dict.update(yaml_dict)

    if config_data_dict:
        env_dict.update(config_data_dict)

    return env_dict


def get_default_gateway(ipv6=False):
    """Returns a tuple with ipaddress and interface of the default gateway.

    :param ipv6: If true return ip6 address and interface else ip4 address
    :raises IptablesError: If there is no default gateway.
    :returns: a tuple with (ip_address, interface) on success or (None, None) on failure
    """
    try:
        default_gateway = netifaces.gateways()["default"]
    except KeyError:
        raise IptablesError("Could not find a default gateway. Are you connected?")

    if ipv6:
        try:
            gateway = default_gateway[netifaces.AF_INET6]
            return gateway
        except KeyError:
            return (None, None)

    try:
        gateway = default_gateway[netifaces.AF_INET]
        return gateway
    except KeyError:
        return (None, None)


def get_interface_addresses(iface):
    """Returns the iface addresses as dictionary of the given interface."""
    iface_addresses = netifaces.ifaddresses(iface)

    default_iface_dict = {"link": iface_addresses[netifaces.AF_LINK]}
    try:
        default_iface_dict["inet"] = iface_addresses[netifaces.AF_INET]
    except KeyError:
        pass

    try:
        default_iface_dict["inet6"] = iface_addresses[netifaces.AF_INET6]
    except KeyError:
        pass

    return default_iface_dict


def render_template(config_file, server=None, protocol=None):
    """Render a jinja2 template with data from config.yml per default. Adds some
    useful variables to the environment which can be uses in rules and fallback
    files.

    :param config_file: the template
    :param server: if None this defaults to 0.0.0.0/0
    :param protocol: if None this defaults to 'udp'
    :returns: the rendered template as string
    """
    config_data_file = resources.get_config_file()
    env = Environment(
        loader=FileSystemLoader(os.path.dirname(config_data_file)),
        trim_blocks=True,
        lstrip_blocks=True,
    )
    with open(config_data_file, "r") as config_data:
        config_data_dict = yaml.safe_load(config_data)
        if server:
            config_data_dict["vpn_remote"] = server["ip_address"]
        else:
            config_data_dict["vpn_remote"] = "0.0.0.0/0"

        if protocol:
            config_data_dict["vpn_protocol"] = protocol
        else:
            config_data_dict["vpn_protocol"] = "udp"
            protocol = "udp"

        if protocol == "udp":
            config_data_dict["vpn_port"] = "1194"
        elif protocol == "tcp":
            config_data_dict["vpn_port"] = "443"
        else:
            raise TypeError("Unknown protocol '{}'.".format(protocol))

        default_ip, default_iface = get_default_gateway(ipv6=False)
        if default_iface:
            try:
                interface_addresses = get_interface_addresses(default_iface)
                config_data_dict["gateway"] = {
                    "ip_address": default_ip,
                    "interface": default_iface,
                }
                config_data_dict["lan"] = interface_addresses
            except KeyError:
                config_data_dict["gateway"] = {}
                config_data_dict["lan"] = {}

        config_data_dict = merge_environment(config_data_dict)
        template = env.get_template(os.path.basename(config_file))
        return template.render(config_data_dict)


@user.needs_root
def read_config(config_file, server=None, protocol=None):
    """High-level abstraction for the render_template method

    :param config_file: the template
    :param server: the server as dict or None
    :param protocol: the used protocol as string. may be one of 'udp' or 'tcp'
    :returns: the rendered template file as dictionary
    """

    rendered_template = render_template(config_file, server, protocol)
    return yaml.safe_load(rendered_template)


class IptablesPrettyFormatter(Formatter):
    """Pretty format for iptables"""

    def format_table_header(self, table, sep="+"):
        """Format the table header

        :param object table: An iptables table
        :param sep: Fill with separator
        :returns: the table surrounded by line filled with sep
        """

        if is_table_v6(table):
            version = "v6"
        else:
            version = "v4"

        prefix = sep * 2
        string = table.name.upper()
        suffix = sep * (self.max_line_length - 5 - len(version) - len(string))

        table_header = "{} {} {} {}".format(prefix, string, version, suffix)
        return table_header

    def format_chain_header(self, chain, sep="="):
        """Format the chain header

        :param object chain: An iptables chain
        :param sep: Fill or separator
        :returns: the chain centered in a filled line with sep
        """
        policy = chain.get_policy()
        if policy:
            policy_s = policy.name
        else:
            policy_s = "None"

        if chain.is_builtin():
            packet_counter, byte_counter = chain.get_counters()
            packet_counter_s = self._format_counter(packet_counter, true_byte=False)
            byte_counter_s = self._format_counter(byte_counter)

            counter_output = "packets: {:>4} bytes: {:>4}".format(
                packet_counter_s, byte_counter_s
            )

            string = "{} ({:^6}) [{}]".format(chain.name, policy_s, counter_output)
        else:
            string = "{} ({:^6})".format(chain.name, policy_s)

        return self.center_string(string, sep)

    @staticmethod
    def _format_iprange(iprange):
        """Format an iprange to cidr notation

        :param iprange: An iprange as string
        :returns: iprange in cidr notation
        """

        if iprange.startswith("!"):
            return "!" + str(netaddr.IPNetwork(iprange.lstrip("!")).cidr)

        return str(netaddr.IPNetwork(iprange).cidr)

    @staticmethod
    def _format_counter(counter, true_byte=True):
        if true_byte:
            multi = 1024
        else:
            multi = 1000

        kilo = multi
        mega = multi * kilo
        giga = multi * mega

        counter_s = ""
        if counter > giga:
            counter_s = str(int(counter / giga)) + "G"
        elif counter > mega:
            counter_s = str(int(counter / mega)) + "M"
        elif counter > kilo:
            counter_s = str(int(counter / kilo)) + "K"
        else:
            counter_s = str(counter)

        return counter_s

    def format_rule(self, rule, rule_number, sep="-"):
        """Format a rule

        :param object rule: An iptables rule
        :param rule_number: the number of the rule in the chain
        :param sep: Separator/Fill
        :returns: the formatted rule in 2 lines with a separating rule appended
        """

        # convert to short cidr notation
        src_net = IptablesPrettyFormatter._format_iprange(str(rule.src))
        dst_net = IptablesPrettyFormatter._format_iprange(str(rule.dst))

        parameters = rule.target.get_all_parameters()
        if parameters:
            parameters_s = str(parameters)
        else:
            parameters_s = ""

        output = "{:3}: {!s:11} {!s:11} {:6} {!s:18} {!s:18} {:<6}{!s}\n".format(
            rule_number,
            rule.in_interface,
            rule.out_interface,
            rule.protocol,
            src_net,
            dst_net,
            rule.target.name,
            parameters_s,
        )

        if rule.matches:
            matches = ""
            for match in rule.matches:
                matches += "{}{!s},".format(match.name, match.parameters)

            output += "     Matches: {}".format(matches.rstrip(","))

        output += "\n"

        packet_counter, byte_counter = rule.get_counters()
        packet_counter = self._format_counter(packet_counter, true_byte=False)
        byte_counter = self._format_counter(byte_counter)

        counter_output = "packets: {:>4} bytes: {:>4}".format(
            packet_counter, byte_counter
        )
        output += "{} [{}] {}".format(
            sep * (self.max_line_length - 6 - len(counter_output)),
            counter_output,
            sep * 2,
        )
        return output


def print_table(table, stream):
    formatter = IptablesPrettyFormatter()
    stream_file = formatter.get_stream_file(stream)

    table.refresh()
    table_header = formatter.format_table_header(table)
    print(table_header, file=stream_file)
    for chain in table.chains:
        chain_header = formatter.format_chain_header(chain)
        print(chain_header, file=stream_file)
        counter = 1
        for rule in chain.rules:
            rule_s = formatter.format_rule(rule, counter)
            print(rule_s, file=stream_file)
            counter += 1

    return formatter.get_output()


@user.needs_root
def to_string(tables=None, version="4", stream=False):
    """Formats tables, chains and rules. If stream is True print directly to
    stdout else collect all lines in formatter.output

    :param tables: list of valid netfilter tables defaults to all netfilter tables
    :param version: either '6' for ip6tables, '4' for iptables or 'all' for both
    :param stream: If true stream to stdout instead of formatter.output
    """

    output = ""
    if version in ("4", "all"):
        if tables is None:
            tables = Table.ALL

        for table_s in tables:
            table = Table(table_s)
            output += print_table(table, stream)

    if version in ("6", "all"):
        if tables is None:
            tables = Table6.ALL

        for table_s in tables:
            table = Table6(table_s)
            output += print_table(table, stream)

    return output


def print_iptables(tables=None, version="4"):
    """Convenience function to print given tables in version to stdout"""
    to_string(tables, version, stream=True)



#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

# connord - connect to nordvpn servers
# Copyright (C) 2019  Mael Stor <maelstor@posteo.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Main module for connord"""

import argparse
import sys
import re
import time
import errno

from requests.exceptions import RequestException

from connord import update
from connord import version
from connord import listings
from connord import connect
from connord import iptables
from connord import user
from connord import servers
from connord import resources
from connord import areas
from connord import countries
from connord import features
from connord import categories
from connord import Printer
from connord import sqlite
from .features import FeatureError


# pylint: disable=too-few-public-methods
class DomainType:

    pattern = re.compile(r"(?P<country_code>[a-z]{2})(?P<number>[0-9]+)(.netflix.com)?")

    def __call__(self, value):
        match = self.pattern.match(value)
        if not match:
            raise argparse.ArgumentTypeError(
                "'{}' is not a valid domain. Expected format is"
                " {{country_code}}{{number}}[.netflix.com]".format(value)
            )

        country_code = match.groupdict()["country_code"]
        CountryType().__call__(country_code)
        return value


class CountryType:
    def __call__(self, value):
        try:
            countries.verify_countries([value])
        except countries.CountryError:
            raise argparse.ArgumentTypeError(
                "'{}' is an unrecognized country.".format(value)
            )
        return value


class AreaType:
    def __call__(self, value):
        try:
            areas.verify_areas([value])
        except (areas.AreaError, ValueError) as error:
            raise argparse.ArgumentTypeError(str(error))

        return value


class CategoryType:
    def __call__(self, value):
        try:
            categories.verify_categories([value])
        except categories.CategoriesError:
            raise argparse.ArgumentTypeError(
                "'{}' is an unrecognized categories.".format(value)
            )

        return value


class FeatureType:
    def __call__(self, value):
        try:
            features.verify_features([value])
        except FeatureError:
            raise argparse.ArgumentTypeError(
                "'{}' is an unrecognized feature.".format(value)
            )

        return value


class LoadType:
    def __call__(self, value):
        try:
            int_value = int(value)
        except ValueError:
            raise argparse.ArgumentTypeError(
                "'{}' must be an integer between 0 and 100.".format(value)
            )

        if int_value < 0 or int_value > 100:
            raise argparse.ArgumentTypeError(
                "'{}' must be a value between 0 and 100.".format(value)
            )

        return int_value


class TopType:
    def __call__(self, value):
        try:
            int_value = int(value)
        except ValueError:
            raise argparse.ArgumentTypeError(
                "'{}' must be a positive integer greater than 0.".format(value)
            )

        if int_value <= 0:
            raise argparse.ArgumentTypeError(
                "'{}' must be a positive integer greater than 0.".format(value)
            )


class TableType:
    def __call__(self, value):
        if iptables.verify_table(value):
            return value

        raise argparse.ArgumentTypeError(
            "'{}' is not a valid table name. Consult man iptables for details.".format(
                value
            )
        )


# pylint: disable=too-many-statements,too-many-locals
def parse_args(argv):
    """Parse arguments

    :returns: list of args
    """
    description = (
        "CønNørD connects you to NordVPN servers with "
        "OpenVPN (https://openvpn.net) and you can choose between a high amount "
        "of possible filters, to make it easy for you to connect "
        "to servers with the best performance and features the server offers to "
        "you. It's taken care that your DNS isn't leaked."
    )
    epilog = "Run a command with -h or --help for more information."
    parser = argparse.ArgumentParser(description=description, epilog=epilog)
    verbosity = parser.add_mutually_exclusive_group()
    verbosity.add_argument(
        "-q", "--quiet", action="store_true", help="Suppress error messages."
    )
    verbosity.add_argument(
        "-v", "--verbose", action="store_true", help="Show what's going."
    )
    command = parser.add_subparsers(dest="command")
    description = (
        "Update NordVPN configuration files and the location "
        "database. Use --force to force an update. Normally 'update' allows for "
        "one update per day."
    )
    help_message = "Update nordvpn configuration files and the location database."
    update_cmd = command.add_parser(
        "update", help=help_message, description=description
    )
    update_cmd.add_argument(
        "-f",
        "--force",
        action="store_true",
        help="Force an update of the configuration files.",
    )
    description = (
        "The 'list' command allows some powerful filtering so you can "
        "select a server tailored to your demands. 'categories', 'features', "
        "'areas', 'countries' give you an overview output of possible "
        "arguments to the respective flag."
    )
    help_message = "List features, categories, ... and servers."
    list_cmd = command.add_parser(
        "list", help=help_message, description=description, epilog=epilog
    )
    list_ipt_parent = argparse.ArgumentParser(description="List iptables")
    list_ipt_parent.add_argument(
        "-4", dest="v4", action="store_true", help="(Default) List ipv4 rules"
    )
    list_ipt_parent.add_argument(
        "-6", dest="v6", action="store_true", help="List ipv6 rules"
    )
    list_all_table = list_ipt_parent.add_mutually_exclusive_group()
    list_all_table.add_argument(
        "-t",
        "--table",
        type=TableType(),
        action="append",
        help="List TABLE. May be specified multiple times.",
    )
    list_all_table.add_argument(
        "-a", "--all", action="store_true", help="List all tables."
    )
    list_cmd_sub = list_cmd.add_subparsers(dest="list_sub")
    list_cmd_sub.add_parser(
        "iptables",
        add_help=False,
        help=(
            "List rules in netfilter tables."
            " With no arguments list ipv4 'filter' table."
        ),
        parents=[list_ipt_parent],
    )
    list_cmd_sub.add_parser(
        "countries", help="List all countries with NordVPN servers."
    )
    list_cmd_sub.add_parser("areas", help="List all areas/cities with NordVPN servers.")
    list_cmd_sub.add_parser(
        "features", help="List all possible features of NordVPN servers."
    )
    list_cmd_sub.add_parser(
        "categories", help="List all possible categories of NordVPN servers."
    )
    description = (
        "List servers filtered by one or more of the specified arguments."
        " If no arguments are given list all NordVPN servers. To see possible "
        "values for --area, --country, --feature and --category have a look at the "
        "respective 'list' commands."
    )
    help_message = "List servers filtered by specified arguments."
    list_servers = list_cmd_sub.add_parser(
        "servers", help=help_message, description=description
    )
    list_servers.add_argument(
        "-c",
        "--country",
        action="append",
        type=CountryType(),
        help="Select a specific country. May be specified multiple times.",
    )
    list_servers.add_argument(
        "-a",
        "--area",
        action="append",
        type=AreaType(),
        help="Select a specific area. May be specified multiple times.",
    )
    list_servers.add_argument(
        "-f",
        "--feature",
        action="append",
        type=FeatureType(),
        help="Select servers with a specific list of features. May be"
        " specified multiple times.",
    )
    list_servers.add_argument(
        "-t",
        "--category",
        action="append",
        type=CategoryType(),
        help="Select servers with a specific category. May be specified multiple"
        " times.",
    )
    list_servers.add_argument(
        "--netflix", action="store_true", help="Select servers configured for netflix."
    )
    list_load_group = list_servers.add_mutually_exclusive_group()
    list_load_group.add_argument(
        "--max-load",
        dest="max_load",
        type=LoadType(),
        help="Filter servers by maximum load.",
    )
    list_load_group.add_argument(
        "--min-load",
        dest="min_load",
        type=LoadType(),
        help="Filter servers by minimum load.",
    )
    list_load_group.add_argument(
        "--load", type=LoadType(), help="Filter servers by exact load match."
    )
    list_servers.add_argument(
        "--top", type=TopType(), help="Show TOP count resulting servers."
    )
    description = (
        "Connect to NordVPN servers. You may specify a single server of "
        "your choice with -s SERVER, where SERVER is a COUNTRY_CODE followed by "
        "a number. This must be a SERVER for which a configuration file exists. "
        "The same filters --area, --country ... like in the 'list' command can "
        "be applied here. For a list of possible values see the respective "
        "command in the 'list' command. "
        "IMPORTANT: Note that currently obfuscated servers aren't supported. "
    )
    connect_cmd = command.add_parser(
        "connect", help="Connect to a server.", description=description
    )
    server_best = connect_cmd.add_mutually_exclusive_group()
    server_best.add_argument(
        "-s",
        "--server",
        type=DomainType(),
        nargs=1,
        help="Connect to a specific server. Only -d, -o, --udp and --tcp have an "
        "effect.",
    )
    server_best.add_argument(
        "-b",
        "--best",
        action="store_true",
        help="Use best server depending on server load. When multiple servers"
        " got the same load use the one with the best ping. This is the default "
        "behaviour if not specified on the command-line",
    )
    connect_cmd.add_argument(
        "-c",
        "--country",
        action="append",
        type=CountryType(),
        help="Select a specific country. May be specified multiple times.",
    )
    connect_cmd.add_argument(
        "-a",
        "--area",
        action="append",
        type=AreaType(),
        help="Select a specific area. May be specified multiple times.",
    )
    connect_cmd.add_argument(
        "-f",
        "--feature",
        action="append",
        type=FeatureType(),
        help="Select servers with a specific feature. May be"
        " specified multiple times.",
    )
    connect_cmd.add_argument(
        "-t",
        "--category",
        action="append",
        type=CategoryType(),
        help="Select servers with a specific category. May be specified multiple"
        " times.",
    )
    connect_cmd.add_argument(
        "--netflix", action="store_true", help="Select servers configured for netflix."
    )
    connect_load_group = connect_cmd.add_mutually_exclusive_group()
    connect_load_group.add_argument(
        "--max-load",
        dest="max_load",
        type=LoadType(),
        help="Filter servers by maximum load.",
    )
    connect_load_group.add_argument(
        "--min-load",
        dest="min_load",
        type=LoadType(),
        help="Filter servers by minimum load.",
    )
    connect_load_group.add_argument(
        "--load", type=LoadType(), help="Filter servers by exact load match."
    )
    connect_cmd.add_argument(
        "-d", "--daemon", action="store_true", help="Start in daemon mode."
    )
    connect_cmd.add_argument(
        "-o",
        "--openvpn",
        dest="openvpn_options",
        type=str,
        nargs=1,
        help="Options to pass through to openvpn as single string",
    )
    udp_tcp = connect_cmd.add_mutually_exclusive_group()
    udp_tcp.add_argument(
        "--udp", action="store_true", help="Use UDP protocol. This is the default"
    )
    udp_tcp.add_argument(
        "--tcp",
        action="store_true",
        help="Use TCP protocol. Only one of --udp or --tcp may be present.",
    )
    description = (
        "Kill the openvpn process spawned by connord or all openvpn "
        "processes with --all"
    )
    kill_cmd = command.add_parser(
        "kill", help="Kill the openvpn process.", description=description
    )
    kill_cmd.add_argument(
        "-a", "--all", action="store_true", help="Kill all openvpn processes."
    )
    description = (
        "Manage your iptables configuration with this command. Under normal "
        "cirumstances your rules files are automatically applied when running "
        "'connect'. "
        "If somethings going wrong or you are modifying  rules you can try one "
        "of the commands defined here. "
        "You can list your current iptables rules with the 'list' command as an "
        "alternative to the native iptables -L -vn --line-num."
    )
    iptables_cmd = command.add_parser(
        "iptables", help="Manage iptables.", description=description, epilog=epilog
    )
    iptables_cmd_subparsers = iptables_cmd.add_subparsers(dest="iptables_sub")
    iptables_cmd_subparsers.add_parser(
        "list", parents=[list_ipt_parent], add_help=False
    )
    help_message = "Reload iptables configuration."
    description = (
        "Reload iptables 'rules' configuration when connected to "
        "NordVPN or else the 'fallback' configuration. Especially useful after "
        "editing a configuration file and you wish to apply it to your running "
        "iptables rules."
    )
    iptables_cmd_subparsers.add_parser(
        "reload", help=help_message, description=description
    )
    help_message = "Flush iptables."
    description = (
        "Flush iptables to fallback configuration or with --no-fallback to no "
        "rules at all and apply ACCEPT policy to builtin chains."
    )
    flush_cmd = iptables_cmd_subparsers.add_parser(
        "flush", help=help_message, description=description
    )
    flush_cmd.add_argument(
        "--no-fallback",
        dest="no_fallback",
        action="store_true",
        help="Flush tables ignoring fallback files.",
    )
    help_message = "Apply iptables rules per 'table'."
    description = (
        "Apply iptables rules defined in 'rules' or 'fallback' "
        "configuration files for a specific 'table'."
    )
    apply_cmd = iptables_cmd_subparsers.add_parser(
        "apply", help=help_message, description=description
    )
    apply_cmd.add_argument(
        "table", type=TableType(), help="Apply iptables rules for 'table'."
    )
    apply_cmd.add_argument(
        "-f",
        "--fallback",
        action="store_true",
        help="Apply fallback instead of the rules configuration.",
    )
    apply_cmd.add_argument(
        "-6", dest="ipv6", action="store_true", help="Apply the ipv6 configuration."
    )
    command.add_parser("version", help="Show version")

    return parser.parse_args(argv)


@user.needs_root
def process_list_ipt_cmd(args):
    if args.v4 and args.v6:
        version_ = "all"
    elif args.v4:
        version_ = "4"
    elif args.v6:
        version_ = "6"
    else:
        version_ = "4"

    if args.all:
        tables = None
    elif args.table:
        tables = args.table
    else:
        tables = ["filter"]

    return listings.list_iptables(tables, version_)


def process_list_servers_cmd(args):
    countries_ = args.country
    area_ = args.area
    categories_ = args.category
    features_ = args.feature
    netflix = args.netflix
    top = args.top

    if args.max_load:
        load_ = args.max_load
        match = "max"
    elif args.min_load:
        load_ = args.min_load
        match = "min"
    elif args.load:
        load_ = args.load
        match = "exact"
    else:  # apply defaults
        load_ = 100
        match = "max"

    return listings.list_servers(
        countries_, area_, categories_, features_, netflix, load_, match, top
    )


def process_list_cmd(args):
    """
    Process arguments when command is 'list'

    :param args: Command-line arguments
    :returns: True if processing was successful
    """

    if args.list_sub == "iptables":
        process_list_ipt_cmd(args)
    elif args.list_sub == "servers":
        process_list_servers_cmd(args)
    elif args.list_sub == "countries":
        listings.list_countries()
    elif args.list_sub == "areas":
        listings.list_areas()
    elif args.list_sub == "features":
        listings.list_features()
    elif args.list_sub == "categories":
        listings.list_categories()
    else:
        # default: list all servers
        listings.list_servers(None, None, None, None, None, 100, "max", None)


@user.needs_root
def process_connect_cmd(args):
    """
    Process arguments for connect command

    :param object args: Command-line arguments
    :returns: True if processing was successful
    """

    if args.server:
        server = args.server
    elif args.best:
        server = "best"
    else:  # apply default
        server = "best"

    countries_ = args.country
    areas_ = args.area
    features_ = args.feature
    categories_ = args.category
    netflix = args.netflix

    if args.max_load:
        load_ = args.max_load
        match = "max"
    elif args.min_load:
        load_ = args.min_load
        match = "min"
    elif args.load:
        load_ = args.load
        match = "exact"
    else:  # apply defaults
        load_ = 10
        match = "max"

    daemon = args.daemon
    if args.openvpn_options:
        openvpn = args.openvpn_options[0]
    else:
        openvpn = args.openvpn_options

    if args.udp:
        protocol = "udp"
    elif args.tcp:
        protocol = "tcp"
    else:  # apply default
        protocol = "udp"

    return connect.connect(
        server,
        countries_,
        areas_,
        features_,
        categories_,
        netflix,
        load_,
        match,
        daemon,
        openvpn,
        protocol,
    )


@user.needs_root
def process_iptables_cmd(args):
    """Process 'iptables' command

    :param object args: command-line arguments as Namespace
    :returns: True if processing was successful
    """

    if args.iptables_sub == "flush":
        if args.no_fallback:
            iptables.reset(fallback=False)
        else:
            iptables.reset(fallback=True)

    elif args.iptables_sub == "apply":
        config_file = iptables.get_config_path(
            args.table, args.fallback, ipv6=args.ipv6
        )

        if args.fallback:
            server = None
            protocol = None
        else:
            try:
                server = resources.get_stats(stats_name="server")
                protocol = resources.get_stats()["last_server"]["protocol"]
            except (resources.ResourceNotFoundError, KeyError):
                raise iptables.IptablesError(
                    "Cannot apply 'rules' files when not connected to a NordVPN server."
                )
        return iptables.apply_config(config_file, server, protocol)
    elif args.iptables_sub == "reload":
        try:
            server = resources.get_stats(stats_name="server")
            protocol = resources.get_stats()["last_server"]["protocol"]
            filetype = "rules"
        except (resources.ResourceNotFoundError, KeyError):
            filetype = "fallback"

        return iptables.apply_config_dir(server, protocol, filetype=filetype)
    elif args.iptables_sub == "list":
        return process_list_ipt_cmd(args)
    else:
        raise NotImplementedError("Not implemented")

    return True


@user.needs_root
def process_kill_cmd(args):
    """Process 'kill' command

    :param object args: Namespace object holding the command-line arguments
    """

    if args.all:
        connect.kill_openvpn()
    else:
        ovpn_pid = resources.read_pid()
        connect.kill_openvpn(ovpn_pid)


# This function has a high complexity score but it's kept simple though
# pylint: disable=too-many-branches
def main():  # noqa: C901
    """Entry Point for the program. A first level argument processing method.
    All Exceptions that lead to an exit of the program are catched here.

    :raises: SystemExit either 0 or 1
    """

    if not sys.argv[1:]:
        sys.argv.extend(["-h"])

    args = parse_args(sys.argv[1:])
    # TODO: recognize configuration in config.yml
    printer = Printer(verbose=args.verbose, quiet=args.quiet)

    try:
        if args.command == "update":
            update.update(force=args.force)
        elif args.command == "version":
            version.print_version()
        elif args.command == "list":
            process_list_cmd(args)
        elif args.command == "connect":
            process_connect_cmd(args)
        elif args.command == "kill":
            process_kill_cmd(args)
        elif args.command == "iptables":
            process_iptables_cmd(args)
        else:
            # This should only happen when someone tampers with sys.argv
            raise NotImplementedError("Could not process command-line arguments.")
        sys.exit(0)
    except PermissionError:
        printer.error(
            'Permission Denied: You need to run "connord {}" as root'.format(
                args.command
            )
        )
    except resources.ResourceNotFoundError as error:
        printer.error(str(error))
    except resources.MalformedResourceError as error:
        printer.error(str(error))
    except areas.AreaError as error:
        printer.error(str(error))
    except iptables.IptablesError as error:
        printer.error(str(error))
    except countries.CountryError as error:
        printer.error(str(error))
    except FeatureError as error:
        printer.error(str(error))
    except servers.DomainNotFoundError as error:
        printer.error(str(error))
    except servers.MalformedDomainError as error:
        printer.error(str(error))
    except connect.ConnectError as error:
        printer.error(str(error))
    except update.UpdateError as error:
        printer.error(str(error))
    except RequestException as error:
        printer.error(str(error))
    except sqlite.SqliteError as error:
        printer.error(str(error))
    except IOError as error:
        # Don't handle broken pipe
        if error.errno != errno.EPIPE:
            raise
    except KeyboardInterrupt:
        time.sleep(0.5)
        sys.exit(0)

    sys.exit(1)



# vim: set fileencoding=utf-8 :

# connord - connect to nordvpn servers
# Copyright (C) 2019  Mael Stor <maelstor@posteo.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""List servers"""

from connord import servers
from connord import countries
from connord import categories
from connord import features
from connord import load
from connord import iptables
from connord import areas


def filter_servers_by_count(servers_, top):
    """
    Filter servers to just show top count results

    :param servers_: List of servers
    :param top: Integer to show count servers
    :returns: The filtered servers
    """

    servers_ = servers_.copy()
    filtered_servers = [server for i, server in enumerate(servers_, 1) if i <= top]
    return filtered_servers


def list_iptables(tables, version):
    """Prints iptables to stdout

    :param tables: list of tables like ['filter']
    :param version: '6' or '4' or 'all'
    :returns: True
    """
    iptables.print_iptables(tables, version)
    return True


def list_countries():
    countries.print_countries()


def list_areas():
    areas.print_areas()


def list_features():
    features.print_features()


def list_categories():
    categories.print_categories()


def filter_servers(
    servers_, netflix, countries_, areas_, features_, categories_, load_, match, top
):
    """High-level abstraction to filter servers by given filters

    :returns: the filtered servers
    """
    servers_ = servers_.copy()
    if load_:
        servers_ = load.filter_servers(servers_, load_, match)
    if netflix:
        servers_ = servers.filter_netflix_servers(servers_, countries_)
    if countries_:
        servers_ = countries.filter_servers(servers_, countries_)
    if areas_:
        servers_ = areas.filter_servers(servers_, areas_)
    if categories_:
        servers_ = categories.filter_servers(servers_, categories_)
    if features_:
        servers_ = features.filter_servers(servers_, features_)
    if top:
        servers_ = filter_servers_by_count(servers_, top)

    return servers_


def list_servers(countries_, areas_, categories_, features_, netflix, load_, match, top):
    """
    List servers filtered by one or more arguments

    :param countries_: List of countries
    :param area_: List of areas
    :param categories_: List of categories
    :param features_: List of features
    :param netflix: If set filter servers optimized for netflix
    :param load_: An integer to filter servers by load.
    :param match: Apply load filter with 'max', 'min' or 'exact' match
    :param top: Show just top count results.
    :returns: True
    """

    servers_ = servers.get_servers()
    servers_ = filter_servers(
        servers_, netflix, countries_, areas_, features_, categories_, load_, match, top
    )

    servers.to_string(servers_, stream=True)
